import { NOTHING, DRAFT_STATE } from './env.js';
import './types-internal.js';
import { isDraftable, die, isDraft, isMap, isSet, each } from './common.js';
import { ImmerScope } from './scope.js';
import { processResult, maybeFreeze } from './finalize.js';
import { createProxyProxy } from './proxy.js';
import { applyPatches, proxyMap, proxySet, createES5Proxy, willFinalizeES5 } from './plugins.js';

/* istanbul ignore next */

function verifyMinified() {}

var configDefaults = {
  useProxies: typeof Proxy !== "undefined" && typeof Proxy.revocable !== "undefined" && typeof Reflect !== "undefined",
  autoFreeze: typeof process !== "undefined" ? process.env.NODE_ENV !== "production" :
  /* istanbul ignore next */
  verifyMinified.name === "verifyMinified",
  onAssign: null,
  onDelete: null,
  onCopy: null
};
var Immer =
/*#__PURE__*/
function () {
  function Immer(config) {
    var _this = this;

    this.useProxies = false;
    this.autoFreeze = false;
    each(configDefaults, function (key, value) {
      var _ref;

      // @ts-ignore
      _this[key] = (_ref = config === null || config === void 0 ? void 0 : config[key]) !== null && _ref !== void 0 ? _ref : value;
    });
    this.setUseProxies(this.useProxies);
    this.produce = this.produce.bind(this);
    this.produceWithPatches = this.produceWithPatches.bind(this);
  }
  /**
   * The `produce` function takes a value and a "recipe function" (whose
   * return value often depends on the base state). The recipe function is
   * free to mutate its first argument however it wants. All mutations are
   * only ever applied to a __copy__ of the base state.
   *
   * Pass only a function to create a "curried producer" which relieves you
   * from passing the recipe function every time.
   *
   * Only plain objects and arrays are made mutable. All other objects are
   * considered uncopyable.
   *
   * Note: This function is __bound__ to its `Immer` instance.
   *
   * @param {any} base - the initial state
   * @param {Function} producer - function that receives a proxy of the base state as first argument and which can be freely modified
   * @param {Function} patchListener - optional function that will be called with all the patches produced here
   * @returns {any} a new state, or the initial state if nothing was modified
   */


  var _proto = Immer.prototype;

  _proto.produce = function produce(base, recipe, patchListener) {
    var _this3 = this;

    // curried invocation
    if (typeof base === "function" && typeof recipe !== "function") {
      var defaultBase = recipe;
      recipe = base;
      var self = this;
      return function curriedProduce(base) {
        var _this2 = this;

        if (base === void 0) {
          base = defaultBase;
        }

        for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
          args[_key - 1] = arguments[_key];
        }

        return self.produce(base, function (draft) {
          var _recipe;

          return (_recipe = recipe).call.apply(_recipe, [_this2, draft].concat(args));
        }); // prettier-ignore
      };
    } // prettier-ignore


    {
      if (typeof recipe !== "function") {
        throw new Error("The first or second argument to `produce` must be a function");
      }

      if (patchListener !== undefined && typeof patchListener !== "function") {
        throw new Error("The third argument to `produce` must be a function or undefined");
      }
    }
    var result; // Only plain objects, arrays, and "immerable classes" are drafted.

    if (isDraftable(base)) {
      var scope = ImmerScope.enter(this);
      var proxy = createProxy(this, base, undefined);
      var hasError = true;

      try {
        result = recipe(proxy);
        hasError = false;
      } finally {
        // finally instead of catch + rethrow better preserves original stack
        if (hasError) scope.revoke();else scope.leave();
      }

      if (typeof Promise !== "undefined" && result instanceof Promise) {
        return result.then(function (result) {
          scope.usePatches(patchListener);
          return processResult(_this3, result, scope);
        }, function (error) {
          scope.revoke();
          throw error;
        });
      }

      scope.usePatches(patchListener);
      return processResult(this, result, scope);
    } else {
      result = recipe(base);
      if (result === NOTHING) return undefined;
      if (result === undefined) result = base;
      maybeFreeze(this, result, true);
      return result;
    }
  };

  _proto.produceWithPatches = function produceWithPatches(arg1, arg2, arg3) {
    var _this4 = this;

    if (typeof arg1 === "function") {
      return function (state) {
        for (var _len2 = arguments.length, args = new Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
          args[_key2 - 1] = arguments[_key2];
        }

        return _this4.produceWithPatches(state, function (draft) {
          return arg1.apply(void 0, [draft].concat(args));
        });
      };
    } // non-curried form

    /* istanbul ignore next */


    if (arg3) die();
    var patches, inversePatches;
    var nextState = this.produce(arg1, arg2, function (p, ip) {
      patches = p;
      inversePatches = ip;
    });
    return [nextState, patches, inversePatches];
  };

  _proto.createDraft = function createDraft(base) {
    if (!isDraftable(base)) {
      throw new Error("First argument to `createDraft` must be a plain object, an array, or an immerable object"); // prettier-ignore
    }

    var scope = ImmerScope.enter(this);
    var proxy = createProxy(this, base, undefined);
    proxy[DRAFT_STATE].isManual = true;
    scope.leave();
    return proxy;
  };

  _proto.finishDraft = function finishDraft(draft, patchListener) {
    var state = draft && draft[DRAFT_STATE];

    if (!state || !state.isManual) {
      throw new Error("First argument to `finishDraft` must be a draft returned by `createDraft`"); // prettier-ignore
    }

    if (state.finalized) {
      throw new Error("The given draft is already finalized"); // prettier-ignore
    }

    var scope = state.scope;
    scope.usePatches(patchListener);
    return processResult(this, undefined, scope);
  }
  /**
   * Pass true to automatically freeze all copies created by Immer.
   *
   * By default, auto-freezing is disabled in production.
   */
  ;

  _proto.setAutoFreeze = function setAutoFreeze(value) {
    this.autoFreeze = value;
  }
  /**
   * Pass true to use the ES2015 `Proxy` class when creating drafts, which is
   * always faster than using ES5 proxies.
   *
   * By default, feature detection is used, so calling this is rarely necessary.
   */
  ;

  _proto.setUseProxies = function setUseProxies(value) {
    this.useProxies = value;
  };

  _proto.applyPatches = function applyPatches$1(base, patches) {
    // If a patch replaces the entire state, take that replacement as base
    // before applying patches
    var i;

    for (i = patches.length - 1; i >= 0; i--) {
      var patch = patches[i];

      if (patch.path.length === 0 && patch.op === "replace") {
        base = patch.value;
        break;
      }
    }

    if (isDraft(base)) {
      // N.B: never hits if some patch a replacement, patches are never drafts
      return applyPatches(base, patches);
    } // Otherwise, produce a copy of the base state.


    return this.produce(base, function (draft) {
      return applyPatches(draft, patches.slice(i + 1));
    });
  };

  return Immer;
}();
function createProxy(immer, value, parent) {
  // precondition: createProxy should be guarded by isDraftable, so we know we can safely draft
  var draft = isMap(value) ? proxyMap(value, parent) : isSet(value) ? proxySet(value, parent) : immer.useProxies ? createProxyProxy(value, parent) : createES5Proxy(value, parent);
  var scope = parent ? parent.scope : ImmerScope.current;
  scope.drafts.push(draft);
  return draft;
}
function willFinalize(immer, scope, thing, isReplaced) {
  if (!immer.useProxies) willFinalizeES5(scope, thing, isReplaced);
}

export { Immer, createProxy, willFinalize };
