import { DRAFT_STATE } from '../env.js';
import { ProxyType } from '../types-internal.js';
import { isDraft, each, createHiddenProperty, shallowCopy, has, is, isEnumerable, assertUnrevoked, isDraftable, latest } from '../common.js';
import { ImmerScope } from '../scope.js';
import '../finalize.js';
import '../proxy.js';
import '../immerClass.js';
import { loadPlugin } from '../plugins.js';

// types only!
function enableES5() {
  function willFinalizeES5(scope, result, isReplaced) {
    scope.drafts.forEach(function (draft) {
      draft[DRAFT_STATE].finalizing = true;
    });

    if (!isReplaced) {
      if (scope.patches) {
        markChangesRecursively(scope.drafts[0]);
      } // This is faster when we don't care about which attributes changed.


      markChangesSweep(scope.drafts);
    } // When a child draft is returned, look for changes.
    else if (isDraft(result) && result[DRAFT_STATE].scope === scope) {
        markChangesSweep(scope.drafts);
      }
  }

  function createES5Proxy(base, parent) {
    var isArray = Array.isArray(base);
    var draft = clonePotentialDraft(base);
    each(draft, function (prop) {
      proxyProperty(draft, prop, isArray || isEnumerable(base, prop));
    });
    var state = {
      type: isArray ? ProxyType.ES5Array : ProxyType.ES5Object,
      scope: parent ? parent.scope : ImmerScope.current,
      modified: false,
      finalizing: false,
      finalized: false,
      assigned: {},
      parent: parent,
      base: base,
      draft: draft,
      copy: null,
      revoked: false,
      isManual: false
    };
    createHiddenProperty(draft, DRAFT_STATE, state);
    return draft;
  } // Access a property without creating an Immer draft.


  function peek(draft, prop) {
    var state = draft[DRAFT_STATE];

    if (state && !state.finalizing) {
      state.finalizing = true;
      var value = draft[prop];
      state.finalizing = false;
      return value;
    }

    return draft[prop];
  }

  function _get(state, prop) {
    assertUnrevoked(state);
    var value = peek(latest(state), prop);
    if (state.finalizing) return value; // Create a draft if the value is unmodified.

    if (value === peek(state.base, prop) && isDraftable(value)) {
      prepareCopy(state); // @ts-ignore

      return state.copy[prop] = state.scope.immer.createProxy(value, state);
    }

    return value;
  }

  function _set(state, prop, value) {
    assertUnrevoked(state);
    state.assigned[prop] = true;

    if (!state.modified) {
      if (is(value, peek(latest(state), prop))) return;
      markChangedES5(state);
      prepareCopy(state);
    } // @ts-ignore


    state.copy[prop] = value;
  }

  function markChangedES5(state) {
    if (!state.modified) {
      state.modified = true;
      if (state.parent) markChangedES5(state.parent);
    }
  }

  function prepareCopy(state) {
    if (!state.copy) state.copy = clonePotentialDraft(state.base);
  }

  function clonePotentialDraft(base) {
    var state = base && base[DRAFT_STATE];

    if (state) {
      state.finalizing = true;
      var draft = shallowCopy(state.draft, true);
      state.finalizing = false;
      return draft;
    }

    return shallowCopy(base);
  } // property descriptors are recycled to make sure we don't create a get and set closure per property,
  // but share them all instead


  var descriptors = {};

  function proxyProperty(draft, prop, enumerable) {
    var desc = descriptors[prop];

    if (desc) {
      desc.enumerable = enumerable;
    } else {
      descriptors[prop] = desc = {
        configurable: true,
        enumerable: enumerable,
        get: function get() {
          return _get(this[DRAFT_STATE], prop);
        },
        set: function set(value) {
          _set(this[DRAFT_STATE], prop, value);
        }
      };
    }

    Object.defineProperty(draft, prop, desc);
  } // This looks expensive, but only proxies are visited, and only objects without known changes are scanned.


  function markChangesSweep(drafts) {
    // The natural order of drafts in the `scope` array is based on when they
    // were accessed. By processing drafts in reverse natural order, we have a
    // better chance of processing leaf nodes first. When a leaf node is known to
    // have changed, we can avoid any traversal of its ancestor nodes.
    for (var i = drafts.length - 1; i >= 0; i--) {
      var state = drafts[i][DRAFT_STATE];

      if (!state.modified) {
        switch (state.type) {
          case ProxyType.ES5Array:
            if (hasArrayChanges(state)) markChangedES5(state);
            break;

          case ProxyType.ES5Object:
            if (hasObjectChanges(state)) markChangedES5(state);
            break;
        }
      }
    }
  }

  function markChangesRecursively(object) {
    if (!object || typeof object !== "object") return;
    var state = object[DRAFT_STATE];
    if (!state) return;
    var base = state.base,
        draft = state.draft,
        assigned = state.assigned,
        type = state.type;

    if (type === ProxyType.ES5Object) {
      // Look for added keys.
      // TODO: looks quite duplicate to hasObjectChanges,
      // probably there is a faster way to detect changes, as sweep + recurse seems to do some
      // unnecessary work.
      // also: probably we can store the information we detect here, to speed up tree finalization!
      each(draft, function (key) {
        if (key === DRAFT_STATE) return; // The `undefined` check is a fast path for pre-existing keys.

        if (base[key] === undefined && !has(base, key)) {
          assigned[key] = true;
          markChangedES5(state);
        } else if (!assigned[key]) {
          // Only untouched properties trigger recursion.
          markChangesRecursively(draft[key]);
        }
      }); // Look for removed keys.

      each(base, function (key) {
        // The `undefined` check is a fast path for pre-existing keys.
        if (draft[key] === undefined && !has(draft, key)) {
          assigned[key] = false;
          markChangedES5(state);
        }
      });
    } else if (type === ProxyType.ES5Array) {
      if (hasArrayChanges(state)) {
        markChangedES5(state);
        assigned.length = true;
      }

      if (draft.length < base.length) {
        for (var i = draft.length; i < base.length; i++) {
          assigned[i] = false;
        }
      } else {
        for (var _i = base.length; _i < draft.length; _i++) {
          assigned[_i] = true;
        }
      } // Minimum count is enough, the other parts has been processed.


      var min = Math.min(draft.length, base.length);

      for (var _i2 = 0; _i2 < min; _i2++) {
        // Only untouched indices trigger recursion.
        if (assigned[_i2] === undefined) markChangesRecursively(draft[_i2]);
      }
    }
  }

  function hasObjectChanges(state) {
    var base = state.base,
        draft = state.draft; // Search for added keys and changed keys. Start at the back, because
    // non-numeric keys are ordered by time of definition on the object.

    var keys = Object.keys(draft);

    for (var i = keys.length - 1; i >= 0; i--) {
      var key = keys[i];
      var baseValue = base[key]; // The `undefined` check is a fast path for pre-existing keys.

      if (baseValue === undefined && !has(base, key)) {
        return true;
      } // Once a base key is deleted, future changes go undetected, because its
      // descriptor is erased. This branch detects any missed changes.
      else {
          var value = draft[key];

          var _state = value && value[DRAFT_STATE];

          if (_state ? _state.base !== baseValue : !is(value, baseValue)) {
            return true;
          }
        }
    } // At this point, no keys were added or changed.
    // Compare key count to determine if keys were deleted.


    return keys.length !== Object.keys(base).length;
  }

  function hasArrayChanges(state) {
    var draft = state.draft;
    if (draft.length !== state.base.length) return true; // See #116
    // If we first shorten the length, our array interceptors will be removed.
    // If after that new items are added, result in the same original length,
    // those last items will have no intercepting property.
    // So if there is no own descriptor on the last position, we know that items were removed and added
    // N.B.: splice, unshift, etc only shift values around, but not prop descriptors, so we only have to check
    // the last one

    var descriptor = Object.getOwnPropertyDescriptor(draft, draft.length - 1); // descriptor can be null, but only for newly created sparse arrays, eg. new Array(10)

    if (descriptor && !descriptor.get) return true; // For all other cases, we don't have to compare, as they would have been picked up by the index setters

    return false;
  }

  loadPlugin("es5", {
    createES5Proxy: createES5Proxy,
    markChangedES5: markChangedES5,
    willFinalizeES5: willFinalizeES5
  });
}

export { enableES5 };
