import '../env.js';
import { ProxyType, Archtype } from '../types-internal.js';
import { die, each, get, has, getArchtype, isMap, isSet } from '../common.js';
import '../scope.js';
import '../finalize.js';
import '../proxy.js';
import '../immerClass.js';
import { loadPlugin } from '../plugins.js';

// Import types only! TODO: force in TS 3.8
function enablePatches() {
  function generatePatches(state, basePath, patches, inversePatches) {
    switch (state.type) {
      case ProxyType.ProxyObject:
      case ProxyType.ES5Object:
      case ProxyType.Map:
        return generatePatchesFromAssigned(state, basePath, patches, inversePatches);

      case ProxyType.ES5Array:
      case ProxyType.ProxyArray:
        return generateArrayPatches(state, basePath, patches, inversePatches);

      case ProxyType.Set:
        return generateSetPatches(state, basePath, patches, inversePatches);
    }
  }

  function generateArrayPatches(state, basePath, patches, inversePatches) {
    var base = state.base,
        assigned = state.assigned,
        copy = state.copy;
    /* istanbul ignore next */

    if (!copy) die(); // Reduce complexity by ensuring `base` is never longer.

    if (copy.length < base.length) {
      var _ref = [copy, base];
      base = _ref[0];
      copy = _ref[1];
      var _ref2 = [inversePatches, patches];
      patches = _ref2[0];
      inversePatches = _ref2[1];
    }

    var delta = copy.length - base.length; // Find the first replaced index.

    var start = 0;

    while (base[start] === copy[start] && start < base.length) {
      ++start;
    } // Find the last replaced index. Search from the end to optimize splice patches.


    var end = base.length;

    while (end > start && base[end - 1] === copy[end + delta - 1]) {
      --end;
    } // Process replaced indices.


    for (var i = start; i < end; ++i) {
      if (assigned[i] && copy[i] !== base[i]) {
        var path = basePath.concat([i]);
        patches.push({
          op: "replace",
          path: path,
          value: copy[i]
        });
        inversePatches.push({
          op: "replace",
          path: path,
          value: base[i]
        });
      }
    }

    var replaceCount = patches.length; // Process added indices.

    for (var _i = end + delta - 1; _i >= end; --_i) {
      var _path = basePath.concat([_i]);

      patches[replaceCount + _i - end] = {
        op: "add",
        path: _path,
        value: copy[_i]
      };
      inversePatches.push({
        op: "remove",
        path: _path
      });
    }
  } // This is used for both Map objects and normal objects.


  function generatePatchesFromAssigned(state, basePath, patches, inversePatches) {
    var base = state.base,
        copy = state.copy;
    each(state.assigned, function (key, assignedValue) {
      var origValue = get(base, key);
      var value = get(copy, key);
      var op = !assignedValue ? "remove" : has(base, key) ? "replace" : "add";
      if (origValue === value && op === "replace") return;
      var path = basePath.concat(key);
      patches.push(op === "remove" ? {
        op: op,
        path: path
      } : {
        op: op,
        path: path,
        value: value
      });
      inversePatches.push(op === "add" ? {
        op: "remove",
        path: path
      } : op === "remove" ? {
        op: "add",
        path: path,
        value: origValue
      } : {
        op: "replace",
        path: path,
        value: origValue
      });
    });
  }

  function generateSetPatches(state, basePath, patches, inversePatches) {
    var base = state.base,
        copy = state.copy;
    var i = 0;
    base.forEach(function (value) {
      if (!copy.has(value)) {
        var path = basePath.concat([i]);
        patches.push({
          op: "remove",
          path: path,
          value: value
        });
        inversePatches.unshift({
          op: "add",
          path: path,
          value: value
        });
      }

      i++;
    });
    i = 0;
    copy.forEach(function (value) {
      if (!base.has(value)) {
        var path = basePath.concat([i]);
        patches.push({
          op: "add",
          path: path,
          value: value
        });
        inversePatches.unshift({
          op: "remove",
          path: path,
          value: value
        });
      }

      i++;
    });
  }

  function applyPatches(draft, patches) {
    patches.forEach(function (patch) {
      var path = patch.path,
          op = patch.op;
      /* istanbul ignore next */

      if (!path.length) die();
      var base = draft;

      for (var i = 0; i < path.length - 1; i++) {
        base = get(base, path[i]);
        if (!base || typeof base !== "object") throw new Error("Cannot apply patch, path doesn't resolve: " + path.join("/")); // prettier-ignore
      }

      var type = getArchtype(base);
      var value = deepClonePatchValue(patch.value); // used to clone patch to ensure original patch is not modified, see #411

      var key = path[path.length - 1];

      switch (op) {
        case "replace":
          switch (type) {
            case Archtype.Map:
              return base.set(key, value);

            /* istanbul ignore next */

            case Archtype.Set:
              throw new Error('Sets cannot have "replace" patches.');

            default:
              // if value is an object, then it's assigned by reference
              // in the following add or remove ops, the value field inside the patch will also be modifyed
              // so we use value from the cloned patch
              // @ts-ignore
              return base[key] = value;
          }

        case "add":
          switch (type) {
            case Archtype.Array:
              return base.splice(key, 0, value);

            case Archtype.Map:
              return base.set(key, value);

            case Archtype.Set:
              return base.add(value);

            default:
              return base[key] = value;
          }

        case "remove":
          switch (type) {
            case Archtype.Array:
              return base.splice(key, 1);

            case Archtype.Map:
              return base.delete(key);

            case Archtype.Set:
              return base.delete(patch.value);

            default:
              return delete base[key];
          }

        default:
          throw new Error("Unsupported patch operation: " + op);
      }
    });
    return draft;
  }

  function deepClonePatchValue(obj) {
    if (!obj || typeof obj !== "object") return obj;
    if (Array.isArray(obj)) return obj.map(deepClonePatchValue);
    if (isMap(obj)) return new Map(Array.from(obj.entries()).map(function (_ref3) {
      var k = _ref3[0],
          v = _ref3[1];
      return [k, deepClonePatchValue(v)];
    }));
    if (isSet(obj)) return new Set(Array.from(obj).map(deepClonePatchValue));
    var cloned = Object.create(Object.getPrototypeOf(obj));

    for (var key in obj) {
      cloned[key] = deepClonePatchValue(obj[key]);
    }

    return cloned;
  }

  loadPlugin("patches", {
    applyPatches: applyPatches,
    generatePatches: generatePatches
  });
}

export { enablePatches };
